import matplotlib as mpl
mpl.use('Agg')

import matplotlib.pyplot as plt
import matplotlib.colors as colors

import sys
import pandas as pd
import numpy as np
#import matplotlib as mpl
import os
import pickle

print(sys.argv[0])

print ('Number of arguments: {} arguments {}'.format( len(sys.argv),sys.argv))

dir_path=sys.argv[1]

#mpl.style.use('seaborn')


networks=["subelj_cora", "cit-HepTh", "wikipedia_link_af", "wikipedia_link_ia", 
          "loc-gowalla_edges", "livemocha", "munmun_twitter_social", "dblp-cite"]

#dir_path = 'save_runs/'
#run_path = networks[3] #4

def sort_fingerprints2(fps):
    cols = []
    assign = []
    conta=0
    init_cols = np.array([i for i in range(len(fps[0]))])
    i=0
    for fp in fps:
        # find non zero elements in the fingerprint
        #c_fp=np.nonzero(fp[init_cols]>0.2)
        c_fp=np.nonzero(fp[init_cols])
        this_cols = init_cols[c_fp]
        cols.extend(this_cols)
        conta+=len(this_cols)
        assign.append(conta-1)
        print(i,end=' ')
        init_cols = np.array([i for i in init_cols if i not in this_cols]).copy()
        i+=1
    return cols, assign


def sort_fingerprints(rows,fmap):
    cols = []
    assign = []
    conta=0
    for r in rows:
        cols.extend(fmap[r])
        conta+=len(fmap[r])
        assign.append(conta-1)
        print(r, end=' ')
    print('|')
    return cols, assign


def plot_fingerprints(dir_path,data,nfp,namefig):
    nr = data.shape[0]
    nc = data.shape[1]

    minc2=min(nfp,nc) # number of positions in fp to display in zoom
    minr2=min(4,nr)  #number of fp to display in zoom

    #fig, ax = plt.subplots() #
    fig = plt.figure(figsize=[12, 4])
    ax = fig.add_subplot(111)

    #ax.matshow(data,cmap=plt.cm.afmhot)
    #ax.matshow(-data,cmap=plt.cm.binary)
    ax.spy(data,markersize=0.5)
    ax.set_aspect('auto')

    extent = [0, nc, 0, nr]
    #axins = ax.inset_axes([0.52, 0.52, .44, .42])
    axins = ax.inset_axes([0.05, 0.15, .44, .42])
    axins.matshow(data, extent=extent,origin="lower",cmap=plt.cm.afmhot)
    
    axins.set_aspect('auto')

    # sub region of the original image
    x1, x2, y1, y2 = 0, minc2, 0, minr2
    axins.set_xlim(x1, x2)
    axins.set_ylim(y1, y2)
    #axins.set_xticklabels('')
    axins.set_yticklabels('')
    axins.invert_yaxis()
    axins.xaxis.tick_bottom()
    axins.grid(color='k', linestyle='-', linewidth=1, axis='y')
    
    ax.indicate_inset_zoom(axins)
    ax.xaxis.tick_bottom()
    
    #plt.title(run_path, fontsize=14)
    plt.ylabel("Centroids")
    plt.xlabel("Trace") # or trace
    #plt.show()
    fig.savefig(dir_path+'fps_'+namefig+'.png')

    
def plot_graph_matrix(dir_path, cols_small, namefig):

    try:
        csr_graph = pickle.load(open(dir_path+'matrix.pkl', "rb"))
    except:
        print("ERROR: cannot load matrix file")

    arr_cs=np.array(cols_small)
    print(type(arr_cs))

    matrix=csr_graph[0].tocsc()[:,arr_cs]
    matrix=matrix[arr_cs,:]

    ncolss=len(cols_small)

    fig, ax = plt.subplots(1, 2, figsize=[12,5])


    ax[0].spy(matrix,precision=0.1,markersize=0.1)
    ax[0].xaxis.tick_bottom()
    ax[0].yaxis.tick_right()
    ax[0].set_xlim(0, ncolss)
    ax[0].set_ylim(0, ncolss)
    ax[0].invert_yaxis()
    ax[0].set_aspect('auto')  

    ax[1].spy(matrix[1:nfp,0:nfp],precision=0.1,markersize=0.1)
    #ax[1].matshow(matrix[1:nfp,0:nfp],cmap=plt.cm.binary,precision=0.1,markersize=1)
    ax[1].xaxis.tick_bottom()
    ax[1].yaxis.tick_right()
    ax[1].set_xlim(0, nfp)
    ax[1].set_ylim(0, nfp)
    ax[1].invert_yaxis()
    ax[1].set_aspect('auto')
    #plt.title('Clustering assignment - '+run_path)
    
    lines = [[[0,ncolss],[i,i]] for i in assign]
    for line in lines:
        ax[0].plot(*line,'r',lw='0.5')

    lines2 = [[[0,nfp],[i,i]] for i in assign if i <= nfp]
    for line in lines2:
        ax[1].plot(*line,'r',lw='1')
    
    fig.savefig(dir_path+'mat_'+namefig+'.png')

    
def load_sort_clusters(dir_path,graph):
    # graph = '10000' 'last' 'clustering'

    if graph == 'last':
        graph_file='temp_last.pkl'
        args=2
    elif graph == 'clustering':
        graph_file='clustering.pkl'
        args=3
    else:
        graph_file='temp_'+graph+'.pkl'
        args=2
    
    if os.path.isfile(dir_path+graph_file):
        print('Loading {} file ...'.format(graph_file))
        if args == 3:
            fps, fmap, fout = pickle.load(open(dir_path+graph_file, "rb"))
        else:
            fps, fmap = pickle.load(open(dir_path+graph_file, "rb"))
    else:
        return -1

    fps_array = np.array(fps)
    print(fps_array.shape)

    rows = np.argsort([-len(fmap[listNodes]) for listNodes in fmap])
    fps_array = fps_array[rows,:].copy()

    cols_small,assign=sort_fingerprints(rows,fmap)
    #cols,assign=sort_fingerprints2(list(fps_array))

    data_plot=fps_array[:, cols_small]    
        
    # plot fingerprints
    if len(assign)>3:
        nfp=assign[3]
    else:
        nfp=assign[-1]  
    
    return data_plot,nfp,cols_small,assign




# log variables  ======================
num_nodes = 0
num_fp = 1
avg_den = 2
avg_time = 3
cum_time = 4
labels = ['Nodes Processed','NUmber of Footprints','Average Density','Processing Time per Node','Time per 100 Nodes']

log = pd.read_csv(dir_path+'log.txt')
log.columns = labels

total_nodes = log.iloc[-1,num_nodes]


indiv =[1,4]
for i in indiv:
    
    df = log.iloc[:,[0,i]]
    fig=plt.figure(figsize=[3,3]);
    df.plot(x=0, y=1); plt.legend(loc='best')
    #plt.title(run_path, fontsize=14)
    #plt.show()
    fig.savefig(dir_path+'perf_'+str(i)+'.png')

    

t = log.iloc[:,[0]]
data1 = log.iloc[:,[2]]
data2 = log.iloc[:,[3]]

fig, ax1 = plt.subplots(figsize=[3,3])

color = 'tab:brown'
ax1.set_xlabel(labels[0])
ax1.set_ylabel(labels[2], color=color)
ax1.plot(t, data1, color=color)
ax1.tick_params(axis='y', labelcolor=color)

ax2 = ax1.twinx()  # instantiate a second axes that shares the same x-axis

color = 'tab:blue'
ax2.set_ylabel(labels[3], color=color)  # we already handled the x-label with ax1
ax2.plot(t, data2, color=color)
ax2.tick_params(axis='y', labelcolor=color)

fig.tight_layout()  # otherwise the right y-label is slightly clipped
#plt.title(run_path, fontsize=14)
#plt.show()
fig.savefig(dir_path+'perf_process.png')



data_plot,nfp,cols_small,assign = load_sort_clusters(dir_path,'clustering')

plot_fingerprints(dir_path,-data_plot,nfp,'clustering')

plot_graph_matrix(dir_path, cols_small,'clustering')

plt.show()

print("NUmber of columns {}".format(len(cols_small)))
print("NUmber of clusters {}".format(len(assign)))



nums = ['10000','20000','30000','40000','50000','last']
for num in nums:
    graph_file = 'temp_'+num+'.pkl'

    if os.path.isfile(dir_path+graph_file):

        print("\n\n======================================================")
        print("Plotting temporary clusters in {}\n".format(graph_file))
        data_plot,nfp,cols_small,assign = load_sort_clusters(dir_path,num)
        print("NUmber of columns {}".format(len(cols_small)))
        print("NUmber of clusters {}".format(len(assign)))
        plot_fingerprints(dir_path,-data_plot,nfp,num)
        plot_graph_matrix(dir_path, cols_small,num)

