import json
from fpc.util import *
import numpy as np
import pandas as pd
from scipy.sparse import csr_matrix
from scipy.sparse import coo_matrix
import networkx as nx
from networkx.algorithms import community 

import cdlib
from cdlib.algorithms import louvain
from cdlib import evaluation

import sys
import re
import urllib
import argparse
import contextlib
import time
import copy
import math
import pickle
from fpc.find_clusters import *


# definition of constants
_ERROR=-1
_SUCCESS=1
_DEBUG=0
_SMALLN=0.000000001


class MetaGraph:
	"""the graph class"""
	def __init__(self,jgraph):
		
		self.number_file = -1
		self.similarity_cluster = 'dotsim'
		self.similarity_merge = 'dotsim'  #nmi
		self.threshold_cluster = 0.2
		self.threshold_merge = 0.15
		self.function_cluster = 'default'
		self.function_merge = 'default'
		self.fps = []
		
		self.time1=0
		self.time2=0
		
		self.g = nx.Graph()
		
		self.jgraph=jgraph
		self.trained=False
		
		self.name=jgraph['name']
		
		self.num_nodes = jgraph['elements_row']+1
		
		if 'url' in jgraph:
			self.url=jgraph['url']
		else:
			self.url=""
			
		if 'separator' in jgraph:
			self.separator=jgraph['separator']
		else:
			self.separator=jgraph[',']
			
		if 'weighted' in jgraph and "TRUE" == jgraph['weighted'].upper():
			self.weighted=True
		else:
			self.weighted=False
			
		if 'directional' in jgraph and "TRUE" == jgraph['directional'].upper():
			self.directional=True
		else:
			self.directional=False
			
		if 'redundant' in jgraph and "TRUE" == jgraph['redundant'].upper():
			self.redundant=True
		else:
			self.redundant=False
			
		if 'ignore' in jgraph:
			self.ignore=jgraph['ignore']
		else:
			self.ignore="#"
		
		# set the fields in their right position
		self.fields = dict()
		cont=0
		for ffields in jgraph['fields']:
			self.fields[ffields.upper()] = cont
			cont += 1
		
		# add weights if the graph is weighted
		if self.weighted:
			if 'weight_min' in jgraph:
				self.minw=float(jgraph['weight_min'])
			else:
				self.minw=0.0
			
			if 'weight_max' in jgraph:
				self.maxw=float(jgraph['weight_max'])
			else:
				self.maxw=1.0
				
			self.rangew=self.maxw-self.minw
			
		self.num_FP=0
		self.csr_FP=csr_matrix([])
				
				
	def print_fields(self):
		print(" Name " + self.name)
		print(" Url " + self.url)
		print(" Separator \'" + self.separator + "\'")
		print(" Ignore character \'"+self.ignore+ "\'")
		if self.weighted:
			print(" Weighted graph")
			print("  Min weight " + str(self.minw))
			print("  Max weight " + str(self.maxw))
			
		if self.directional:
			print(" Directional graph")
		else:
			print(" Symetric graph")
			
			if self.redundant:
				print("   Redundant")
			else:
				print("   Not redundant")
			
		print(" Field FROM  " + str(self.fields['FROM'])) 
		print(" Field TO  " + str(self.fields['TO']))
		print("\n Parameters run:")
		print("   Function cluster: {}".format(self.function_cluster))
		print("   Similarity cluster: {}".format(self.similarity_cluster))
		print("   Threshold cluster: {}".format(self.threshold_cluster))
		print("   Function merge: {}".format(self.function_merge))
		print("   Similarity merge: {}".format(self.similarity_merge))
		print("   Threshold merge: {}".format(self.threshold_merge))

		
		
		
	def check_storage(self,store_path):
		if store_path:
			if is_writable(store_path):
				self.store=store_path
				return(True)
		return(False)


		
	def predict(self,inputs):
		listp=[]
					
		maxp=0;
		whatp=''
		return whatp
				
				
				
	def get_confidence(self):  #needs to be implemented
		return 1
		
		

	def dump_graph(self,file_out):
		
		f = open(file_out, 'w');
		
		f.write("{ \"graph\": {\n");
				
		f.write("} }\n");
		f.close()


	# use the input_file to train the model  # build_graph
	def build_graph(self, tfile):
		#global outputs
		has_trained=0
		cont=0
		# first read the format file
		rowf = -1
		rowg, colg, data = [],[],[]
		nodes = []
		
		if self.weighted: # read a weightherd graph 
			usecolumns = [self.fields['FROM'], self.fields['TO'], self.fields['WEIGHT']]
		else:
			usecolumns = [self.fields['FROM'], self.fields['TO']]
	
		
		if self._PRINTLEV > 1:
			print("READING FILE ",tfile)
				
		#df_graph = pd.read_csv(tfile,sep=self.separator,comment=self.ignore, usecols=usecolumns, dtype={0: 'int64',1:'int64'}, converters={0 : convertNAN, 1 : convertNAN})
		df_graph = pd.read_csv(tfile,sep=self.separator,comment=self.ignore, usecols=usecolumns, dtype={0: 'float64',1:'float64'}, keep_default_na='False')
		if self._PRINTLEV > 2:
			print(df_graph.head())
		int_cols = ['1', '2']
		df_graph[int_cols] = df_graph[int_cols].fillna(0)
		df_graph[int_cols] = df_graph[int_cols].apply(pd.to_numeric, downcast='integer')


		if self._PRINTLEV > 1:
			print("PROCESSING FILE ")
			
		rowg = df_graph.iloc[:,0].tolist()
		colg = df_graph.iloc[:,1].tolist()
		
		self.num_nodes=int(max(max(colg),max(rowg)))+1
		
		nodes = [i for i in range(self.num_nodes)]
		
		if self.weighted:
			data_temp=df_graph.iloc[:,2].tolist() # normalize this
			data = [(float(i)-self.minw)/self.rangew for i in data_temp] # normalize the weighted data
		else:
			data=[1.0] * len(colg)
			
			
			
		if not self.directional and not self.redundant:
			if self._PRINTLEV > 1:
				print("Mirroring edges")
			#edges need to be mirrored
			row_temp=rowg.copy()
			rowg.extend(colg)
			colg.extend(row_temp)
			data.extend(data)
			
			
		if self._PRINTLEV > 2:
			print("Size of nodes {}, col {}, data {}".format(len(nodes),len(colg),len(data)))
			print("Min node {}, Max node {}".format(min(nodes),max(nodes)))
			print("Min col {}, Max col {}".format(min(colg),max(colg)))
			print("Min data {}, Max data {}".format(min(data),max(data)))
		
		
		if self.threshold_cluster <= 0:  # use ad hoc threshold, based on the density of the graph
			den=max(len(colg)/len(nodes),1.1)
			self.threshold_cluster=math.log10(den)/50 #15
			if self._PRINTLEV > 0:
				print("Adaptive threshold for clustering set to: {}".format(self.threshold_cluster))
		
		if self.threshold_merge <= 0:  # use ad hoc threshold, based on the density of the graph
			self.threshold_merge=self.threshold_cluster * 1.5  # 1.5 .75 % of the clustering threshold
			if self._PRINTLEV > 0:
				print("Adaptive threshold for merging set to: {}".format(self.threshold_merge))
				
		
		if self._PRINTLEV > 1:
			print("BUILDING NETWORKX GRAPH ")
		
		self.g.clear()
		self.g.add_nodes_from(nodes)
		
		if self.weighted:
			edges = [(x,y,z) for x, y, z in zip(rowg, colg, data)]
			self.g.add_weighted_edges_from(edges)
		else:
			edges = [(x,y) for x, y in zip(rowg, colg)]
			self.g.add_edges_from(edges)
			
		#nodes = self.g.nodes()
		
		
	
		# create sparse matrix representation of batch
		if self._PRINTLEV > 1:
			print("CREATING CSR MATRIX ")
					
		csr_graph = csr_matrix((data, (rowg, colg)), shape=(self.num_nodes, self.num_nodes))
		with open(self.store+'matrix.pkl', 'wb') as f:  # Python 3: open(..., 'wb')
			pickle.dump([csr_graph], f, protocol=-1)
		
		#process the graph, build fingerprints and determine cluster
		if self._PRINTLEV > 1:
			print("PROCESSING GRAPH ")
		fps, fmap, fmap_out=self.process_graph(csr_graph, nodes)
		
		with open(self.store+'clustering.pkl', 'wb') as f:  # Python 3: open(..., 'wb')
			pickle.dump([fps, fmap, fmap_out], f, protocol=-1)
            
		#evaluate the graph
		if self._PRINTLEV > 1:
			print("EVALUATING GRAPH ")
		try:
			scores=self.evaluate_graph(fmap)        
			with open(self.store+'scores.pkl', 'wb') as f:  # Python 3: open(..., 'wb')
				pickle.dump([scores], f, protocol=-1)
		except:
			print("Cannot compute scores")
            
		#restart variables and counters
		cont=0
		row, col, data = [],[],[]
		more_data=False
		
		if self._PRINTLEV > 1:
			print("-------------------------------\n")
			
		return has_trained
	
		
		
		
			
	def process_graph(self, csr_graph, nodes):
			
		if self._PRINTLEV > 5:
			print("print graph input")
			#print(csr_graph)
			#print(nodes)
			
		# find the initial set of clusters  ======================
		start = time.time()
		if self.function_cluster == 'default':
			fps, fmap, fmap_out = findClusters(nodes, csr_graph, self.fps, self.store, similarity=self.similarity_cluster, threshold=self.threshold_cluster)
		else:
			fps, fmap, fmap_out = findClusters(nodes, csr_graph, self.fps, self.store, similarity=self.similarity_cluster, threshold=self.threshold_cluster)
		
		
		end = time.time()
		self.time1=end-start
		if self._PRINTLEV > 0:
			print("Time to compute clusters: {}".format(end - start))
			print('Clusters found: ' + str(len(fmap)))
		
		with open(self.store+'temp_last.pkl', 'wb') as f:  
			pickle.dump([fps, fmap], f, protocol=-1)
			
		# merge similar clusters =================================
		start = time.time()
		if self.function_merge == 'default':
			merged_fps, merged_fmap = mergeFingerprints(fps, fmap, 'add', self.store, similarity=self.similarity_merge, threshold=self.threshold_merge)
		else:
			merged_fps, merged_fmap = mergeFingerprints(fps, fmap, 'add', self.store, similarity=self.similarity_merge, threshold=self.threshold_merge)
		
		# save the merged fingerprints in the graph structure
		self.fps = merged_fps.copy()
		
		end = time.time()
		self.time2=end-start
		if self._PRINTLEV > 0:
			print("Time to merge clusters: {}".format(end - start))
			print('Clusters merged: ' + str(len(fmap)-len(merged_fmap)))
			print('Remaining clusters: ' + str(len(merged_fmap)))
			
		return merged_fps, merged_fmap, fmap_out  # return the clustering assignment
		
		
		
	def evaluate_graph(self, fmap):
		
		method = 'nidia'
		clusters = list(fmap.values())
		c = cdlib.classes.node_clustering.NodeClustering(clusters, self.g, method)

		# modularity fitness functions
		newman_modularity = evaluation.newman_girvan_modularity(self.g,c).score
		erdos_renyi_mod = evaluation.erdos_renyi_modularity(self.g,c).score
		link_modularity = evaluation.link_modularity(self.g,c).score

		# other fitness_functions
		average_internal_degree = evaluation.average_internal_degree(self.g,c).score
		conductance = evaluation.conductance(self.g,c).score
		cut_ratio = evaluation.cut_ratio(self.g,c).score
		edges_inside = evaluation.edges_inside(self.g,c).score
		internal_edge_density = evaluation.internal_edge_density(self.g,c).score   
		
		scores = defaultdict(dict)
		scores[method]['initial_threshold'] = self.threshold_cluster
		scores[method]['merging_threshold'] = self.threshold_merge
		scores[method]['time_cluster'] = self.time1
		scores[method]['time_merge'] = self.time2
		scores[method]['clusters_found'] = len(fmap)
		#scores[method]['clusters_merged'] = len(fmap)-len(merged_fmap)
		#scores[method]['remaining_clusters'] = len(merged_fmap)

		scores[method]['newman_modularity'] = newman_modularity
		scores[method]['erdos_renyi_mod'] = erdos_renyi_mod
		scores[method]['link_modularity'] = link_modularity
		scores[method]['average_internal_degree'] = average_internal_degree
		scores[method]['conductance'] = conductance
		scores[method]['cut_ratio'] = cut_ratio
		scores[method]['edges_inside'] = edges_inside
		scores[method]['internal_edge_density'] = internal_edge_density
		with open('publish/' + self.name + '_scores'+ '.pkl', 'wb') as f:
			pickle.dump(scores, f)
		
		
		if self._PRINTLEV > 0:
			print("Newman Modularity (max):\t{}".format(newman_modularity))
			print("Erdos Renyi Modularity: \t{}".format(erdos_renyi_mod))
			print("Link Modularity: \t\t{}".format(link_modularity))
			print("Average internal degree: \t{}".format(average_internal_degree))
			print("Conductance (min):\t\t{}".format(conductance))
			print("Edges inside: \t\t\t{}".format(edges_inside))
			print("internal edge density: \t\t{}".format(internal_edge_density))
			
		return scores
		
		
		
	def dump_model(self,dump_model,sufix):
		# use the tree method to dump the model
		self.dump_graph(dump_model+sufix)
		
		return dump_model+sufix
				
				
				
	def process_input(self,tfile,ofile,dumpdata,dumpmodel,url_file,url_path,node_ip, last_trained,last_dump):
		
		last_trained=last_trained+self.build_graph(tfile)
		# concatenate new data into a temp local file
        
		if False:
		# ignore this block for the moment, it is nof fully implemented
			concatenate_files(dumpdata,tfile)
		
			ts=time.time()
			if last_trained > 50 or (ts-last_dump > 600):
				# dump the model in json format
				sufix=str(time.time())
				model_name=self.dump_model(dumpmodel,sufix)
				# dump the data in the same format it was received
				data_name=dump_data(dumpdata,sufix)
				# save the url message to pass to the communication layer
				save_url_message(url_file,url_path,model_name,data_name,last_trained,node_ip,sufix,cidx)
				
				last_trained=0
				last_dump= time.time()

		return last_trained
						


		

	def process_message(self,LOCAL_MESSAGE,input_file):
		# parse the message file
		fmessage = open(LOCAL_MESSAGE, 'r')
		jdata = json.load(fmessage)
		pprint(jdata)
		nattempts=0
		for message in jdata:
		    # now message is a dictionary
		
			if 'model_file' in message: # get the model from the url and add it to the local model
				#first check confidence and timestamp to decide whether to update or not
				mfile=get_remote_file(message["model_file"])
				if mfile != _ERROR:
					self.integrate_model(mfile)
				
			if 'data_file' in message: # get the data file from the url and append it to the local input data
				dfile=get_remote_file(message["data_file"])
				if dfile != _ERROR:
					while(nattempts<10): #make 10 attempts at appending the data into the input file
						nattempts+=1
						if lock_file(input_file)== _SUCCESS:
							#append dfile into Input_file
							fd=open(input_file,'a')
							fd.write(dfile.read()) #NOTE: check what is the actual format of dfile
							fd.close()
							unlock_file(args.input_file)
							break;
						else:
							if self._PRINTLEV>0:
								print("ERROR cannot append data to the local input file ")
							time.sleep(1)
					
				
				

				
	def process_query(self,qfile,ofile):
		try:
			fquery = open(qfile, 'r');
			if self._PRINTLEV > 3:
				print("Opening query file")
				
		except IOError:
			print("Error: Query file cannot be opened")
			return _ERROR
			
		try:
			f = open(ofile, 'a');
			if self._PRINTLEV > 3:
				print("Opening output file")
				
		except IOError:
			print("Error: Output file cannot be opened")
			return _ERROR
			

		isout=False
		
		# first read the format file
		for line in fquery:
			line = line.strip()
			# Read the data and use it for training and/or prediction
			if self._PRINTLEV > 3:
				print("LINE ",line)
				
			inputs=[]  ######################################## Make this a sparse matrix
			
			toks=list(line.split(','))
			if len(toks)>num_params:
					out=toks[num_params];
					isout=True
			for k in range(num_params):
				if upper[k]>lower[k] and toks[k]!="?":
					inputs.append(float(toks[k]))
				else:
					inputs.append(0.0)
					
			# make sure that the query is not empty
			if len(toks)>=num_params:

				# WRITE IN OUTPUT
				if isout:
					f.write(out);
				else:
					f.write(line);
					
				f.write(',');
				
				# Make a prediction
				if is_numeric_class==1:
					p = self.predict(inputs)
					f.write(str(p));

				
				f.write('\n');
				
			
		f.close();
		fquery.close();
		
		return _SUCCESS
	

# end of functions ---------------------






