#import networkx as nx
import scipy as sp
from scipy import sparse
from collections import defaultdict
from scipy.sparse import csr_matrix, spmatrix
import numpy as np
from sklearn.metrics import *
from scipy import spatial
from scipy.sparse import csr_matrix
import scipy.sparse
import scipy as sp
import pickle
import os
import time
import warnings
warnings.filterwarnings('ignore')




def dotSimilarity(fp, vec):
    ''' gets similarity between a fingerprint and a row vector
        the number of non-zero components they share 
        divided by the total number of non-zero components of the vector '''
    return vec.dot(fp).max() / vec.sum()

def jaccardSimilarity(fp, vec):
    ''' gets similarity between a fingerprint and a row vector
        the number of non-zero components they share 
        divided by the total number of non-zero components of the vector '''
    return vec.dot(fp).max() / np.add(vec,fp).sum()

def cosineSimilarity(fp,vec):
    '''' gets similarity between a fingerprint and a row vector'''
    if scipy.sparse.isspmatrix_csr(vec):
        return 1 - spatial.distance.cosine(fp, vec.A.astype(np.float))
    else:
        return 1 - spatial.distance.cosine(fp, vec)

def NMISimilarity(fp,vec):
    '''' gets similarity between a fingerprint and a row vector using NMI'''
    if scipy.sparse.isspmatrix_csr(vec):       
        return normalized_mutual_info_score(fp,vec.A.astype(np.float).flatten())
    else:
        return normalized_mutual_info_score(fp, vec)

def updateFingerprint(fp, vec, countfp, countvec=1):
    ''' updates a fingerprint when a node vector is added to the cluster
        weighted merge of the node vector with the fingerprint '''
    if countvec==1:
        count_total=min(countfp+countvec,1000);
        propfp=min(countfp,999)/count_total
    else:
        count_total=countfp+countvec;
        propfp=countfp/count_total
        
    propvec=countvec/count_total
    
    if isinstance(vec, spmatrix):
        new_fp = fp * propfp + (vec.A.astype(np.float) * propvec)
    else:
        new_fp = fp * propfp + (vec*propvec)
        
    # clip the values so that the fp does not vanish    
    idc=np.nonzero(new_fp>=0.1)
    new_fp[idc]=np.clip(new_fp[idc], 0.6, 1.0) 
    idc=np.nonzero(new_fp)
    new_fp[idc]=np.clip(new_fp[idc], 0.05, 1.0)
        
    return(new_fp)
                

def updateFingerprint_old(fp, vec, countfp, countvec=1):
    ''' updates a fingerprint when a node vector is added to the cluster
        weighted merge of the node vector with the fingerprint '''

    count_total=countfp+countvec;
    propfp=countfp/count_total
        
    propvec=countvec/count_total
    
    if isinstance(vec, spmatrix):
        new_fp = fp * propfp + (vec.A.astype(np.float) * propvec)
        return(new_fp)
    else:
        new_fp = fp * propfp + (vec*propvec)
        return(new_fp)

    
    
# def findProbabilisticClusters(nodes, csr_matrix, fps =[], similarity='dotsim', threshold=0.5):
#     ''' 
#     input
#         nodes: list of nodes
#         csr_matrix: sparse adjacency matrix
#         threshold: initial merging threshold
        
#     returns:
#         fps:  list of all found fingerprints
#         fmap: (dict) fingerprint mapping of nodes to fingerprints 
#               to keep track of what node belongs to what fp
#         {
#             fp_index: [
#                 row_index,
#                 ...
#             ],
#             ...
#         }
#         '''
#     fmap = defaultdict(list)

#     ''' fingerprints '''
#     #fps = []

#     for node in nodes:
#         ri=node
#         row = csr_matrix[ri]
        
#         # initialize fingerprints
#         if len(fps) == 0:
#             fmap[len(fps)].append(node)
#             fps.append(row.A[0].astype(np.float))
#             continue
        
#         # get best scoring fingerprint using dotSimilarity
#         if similarity == 'dotsim':        
#             # sorted and pop gets me the best scoring one (find something more elegant!)
#             similarfps = sorted([(dotSimilarity(fp, row), fi, fp) for fi, fp in enumerate(fps)])

#         # get best scoring fingerprint using cosine Similarity
#         elif similarity == 'cosine':
#             similarfps = sorted([(cosineSimilarity(fp, row), fi, fp) for fi, fp in enumerate(fps)])
        
#         # get best scoring fingerprint using mutual_info_score
#         elif similarity == 'nmi':
#             similarfps = sorted([(NMISimilarity(fp, row), fi, fp) for fi, fp in enumerate(fps)])  

            
#         score, fi, fp = similarfps.pop() # pop the last one
#         total_score=score
#         not_assigned=True  # use this to assign a node only to one cluster
            
#         if score < threshold:  # if less threshold, add as a new fingerprint
#             fmap[len(fps)].append(node)
#             fps.append(row.A[0].astype(np.float))
#         else:
#             for k in reversed(range(len(similarfps))):
#                 if similarfps[k][0] >= threshold: 
#                     total_score+=similarfps[k][0]
#                 else:
#                     break
        
#         while score > threshold:  # otherwise, pop as long as score > threshold and add to other fingerprints proportionally
            
#             proportion=score/total_score
#             # map node to fingerprint
#             if not_assigned: # use this flag to assign a node only to one cluster, remove for multiples
#                 fmap[fi].append(node)
#                 not_assigned=False
                
#             # update fingerprint with row weights
#             fp[:] = updateFingerprint(fp, row*proportion, len(fmap[fi]))
            
#             score, fi, fp = similarfps.pop()
            
            
#     return fps, fmap



def findClusters(nodes, csr_matrixg, fps, save_path, similarity='dotsim', threshold=0.5):
    ''' 
    input
        nodes: list of nodes
        csr_matrix: sparse adjacency matrix
        threshold: initial merging threshold
        
    returns:
        fps:  list of all found fingerprints
        fmap: (dict) fingerprint mapping of nodes to fingerprints 
              to keep track of what node belongs to what fp
        {
            fp_index: [
                row_index,
                ...
            ],
            ...
        }
        '''
    
    _MIN_LINKS = 4   # minimum number of links to be added as a fingerprint
    _LOG_INTERVAL = 100
    _CHECKPOINT_INTERVAL = 10000
    _FLUSH_INTERVAL = 2002
    FLAG_MERGE = False
    OUTLIER_ACTION='remove'
    c_time=0
    cont=0
    c_den=0
    
    fmap = defaultdict(list)
    fmap_oultiers = defaultdict(list)

  
    start_timer=time.time()
    
    for ri,node in enumerate(nodes):
        #ri = node
       
        row = csr_matrixg[ri]
        row_array=row.A[0].astype(np.float)
        
        idxs=np.nonzero(row_array)[0]
        sum_nodes = row_array[idxs].sum()
        if sum_nodes==0: continue
        # book keeping
        cont+=1
        c_den+=sum_nodes  
         
        # initialize fingerprints
        if len(fps) == 0:
            fmap[0].append(node)
            fps.append(row_array)
            array_fps=np.transpose(np.array(fps))
            continue
            
        # compute similarity ====================================================    
        start = time.time() #time the similarity metric
        
        # get best scoring fingerprint using dotSimilarity
        if similarity == 'dotsim':        
            result=row_array[idxs].dot(array_fps[idxs,:])  / sum_nodes
            score=np.amax(result)
            fi=np.where(result == np.amax(result))[0][-1]
          
        # get best scoring fingerprint using cosine Similarity
        elif similarity == 'cosine':
            score, fi, fp = sorted([(cosineSimilarity(fp, row), fi, fp) for fi, fp in enumerate(fps)]).pop() 
        
        # get best scoring fingerprint using mutual_info_score
        elif similarity == 'nmi':
            score, fi, fp = sorted([(NMISimilarity(fp, row), fi, fp) for fi, fp in enumerate(fps)]).pop()  

            
        end=time.time() # end of timer for the similarity metric
        c_time+=end-start
        
        # keep track of status and perform bookkepping==========================
        
        if node % _LOG_INTERVAL == 0: #print and save log every 100 nodes
            print("node {}/{}, size of fps {}, avg density {:.2f}, avg time: {:.6f}, time since {:.6f}  \n".format(node, len(nodes), len(fps), c_den/cont, c_time/cont, end-start_timer))
            with open(save_path+'log.txt', 'a+') as f:
                f.write("{},{},{:.3f},{:.6f},{:.6f}\n".format(node, len(fps), c_den/cont, c_time/cont, end-start_timer))
            start_timer = time.time()
            c_time=0
            cont=0
            c_den=0
              
        if node % _CHECKPOINT_INTERVAL == 0:  #save progress every 10000 nodes
            FLAG_MERGE = True
            OUTLIER_ACTION='add'  # add back the outliers and merge
            with open(save_path+'temp_'+str(node)+'.pkl', 'wb') as f:  # Python 3: open(..., 'wb')
                pickle.dump([fps, fmap], f, protocol=-1)
            
        if node % _FLUSH_INTERVAL == 0:  # remove outliers and merge every 2002 steps
            FLAG_MERGE = True
            OUTLIER_ACTION='remove'
        
        
        # update fingerprint ==================================================
        if score > threshold: 
            # map node to fingerprint
            fmap[fi].append(node)
            # update fingerprint with row weights  REVISE this in other versions, it wasn;t being updated
            fps[fi] = updateFingerprint(fps[fi], row_array, len(fmap[fi])).copy()
            
            
        elif sum_nodes > _MIN_LINKS: #add a new fingerprint and merge if needed
            fmap[len(fps)].append(node)
            fps.append(row_array)
            
            if FLAG_MERGE and len(fps)>200: # 300
                fps_temp, fmap_temp = mergeFingerprints(fps, fmap, OUTLIER_ACTION, save_path, similarity, threshold*.75)
                fmap.clear()
                fmap.update(fmap_temp) #update the current dictionary
                fps.clear()
                fps = fps_temp.copy()
                FLAG_MERGE=False
                
            array_fps=np.transpose(np.array(fps))
            
        else: # node has very few links to be added as a fingerprint, assign to outliers cluster
            fmap_oultiers[0].append(node)
            
            
    return fps, fmap, fmap_oultiers





def mergeFingerprints(fps_temp, fmap_temp, outliers, save_path, similarity='dotsim', threshold=0.3):
    '''
    finds similar clusters and merges them  
    '''
    print("MERGE FINGERPRINTS {}".format(outliers))
    _MIN_DENSITY = 4 # minimum number of elements in cluster to not be flushed
    fmap = {}
    fps=[]
    # same kind of mapping of nodes to fingerprints
    merged_fps = []
    merged_fmap = {}
    
    if outliers == 'remove' or outliers== 'add':
        try:
            with open(save_path+'outliers.pkl', 'rb') as f: 
                out_fps, out_fmap = pickle.load(f)
        except:
            out_fps = []
            out_fmap = {}         
                
    nmore=0
    if outliers == 'add':      
        for i, fp in enumerate(out_fps):
            fmap[i] = out_fmap[i]
            fps.append(out_fps[i])
            nmore+=1
        out_fps.clear() #to empty the file
        out_fmap.clear()
        
        if nmore>0:
            print("# Adding back {} fingerprints".format(nmore))
            os.remove(save_path+'outliers.pkl')
            
    
    # sort fingerprints from less to more density
    idxFP = np.argsort([ len(fmap_temp[listNodes]) for listNodes in fmap_temp])
    # sort fingerprints and change index to map
    for i in range(len(idxFP)):
        fmap[i+nmore] = fmap_temp[idxFP[i]]
        fps.append(fps_temp[idxFP[i]])
    
    
    row_array=np.array(fps)
    # normalize?
    #id_cols=np.nonzero(sum(row_array))
    #row_array[:,id_cols] = row_array[:,id_cols] / row_array[:,id_cols].sum(axis=0) # normalize per column
    #print("shapes {} after {}".format(row_array.shape,row_array.shape))
    #print("sums {} after {}".format(sum(row_array),sum(row_array)))
    
    array_fps=np.transpose(row_array)
    
    processed = []
    for ai, afp in enumerate(fps):
        
        idxs=np.nonzero(row_array[ai,:])[0]
        sum_sonzeros=row_array[ai,idxs].sum()
        #print("Merge fp: {}".format(ai))
        
        # skip fingerprints that have already been merged
        if ai in processed: continue

        # dot similarity
        if similarity == 'dotsim':
            result=row_array[ai,idxs].dot(array_fps[idxs,:])  / sum_sonzeros
            result[ai]=0.0 #make 0 it's own product
            score=np.amax(result)
            bi=np.where(result == np.amax(result))[0][-1]
              
        # normalized mutual info score
        elif similarity == 'nmi':
            score, bi = sorted([(NMISimilarity(afp, bfp), bi) for bi, bfp in enumerate(fps) if bi != ai]).pop()
     
        # cosine Similarity
        elif similarity == 'cosine':
            score, bi = sorted([(cosineSimilarity(afp, bfp), bi) for bi, bfp in enumerate(fps) if bi != ai]).pop() 
                     
                
        # same for second fingerprint
        if bi in processed: continue

        if score > threshold:
            # merge fingerprints
            fps[bi] = updateFingerprint(afp, fps[bi], len(fmap[ai]), len(fmap[bi])).copy() # merge in proportion of densities
            fmap[bi] = list(set(fmap[bi] + fmap[ai]))
            # mark as processed
            processed += [ai]
            
            
    # add fingerprints that were not merged
    allnodes=0
    for i, fp in enumerate(fps):
        
        if outliers=='remove' and len(fmap[i]) <= _MIN_DENSITY: is_outlier = True 
        else: is_outlier = False
        
        if i not in processed and not is_outlier:
            merged_fps.append(fp)
            merged_fmap[len(merged_fps) - 1] = list(fmap[i])
            allnodes+=len(fmap[i])
            
        elif i not in processed and is_outlier:
            out_fps.append(fp)
            out_fmap[len(out_fps) - 1] = list(fmap[i])
            
            
    if outliers == 'remove' and len(out_fps)>0:
        with open(save_path+'outliers.pkl', 'wb') as f:  
            pickle.dump([out_fps, out_fmap], f, protocol=-1)
            print("# Fingerprints removed: {}".format(len(out_fps)))

    #print("Num nodes {}".format(sum([len(merged_fmap[listNodes]) for listNodes in merged_fmap])))
            
    print("# Nodes in fmap: {}\n".format(allnodes))
    return merged_fps, merged_fmap







# def mergeFingerprints_old(fps, fmap, similarity='dotsim', threshold=0.3):
#     '''
#     finds similar clusters and merges them  
#     '''
#     # same kind of mapping of nodes to fingerprints
#     merged_fps = []
#     merged_fmap = {}

#     processed = []
#     for ai, afp in enumerate(fps):
#         # skip fingerprints that have already been merged
#         if ai in processed: continue
            
#         # dot similarity
#         if similarity == 'dotsim':
#             score, bi, bfp = sorted([(dotSimilarity(afp, bfp), bi, bfp) for bi, bfp in enumerate(fps) if bi != ai]).pop()

#         # normalized mutual info score
#         elif similarity == 'nmi':
#             score, bi, bfp = sorted([(NMISimilarity(afp, bfp), bi, bfp) for bi, bfp in enumerate(fps) if bi != ai]).pop()
     
#         # cosine Similarity
#         elif similarity == 'cosine':
#             score, bi, bfp = sorted([(cosineSimilarity(afp, bfp), bi, bfp) for bi, bfp in enumerate(fps) if bi != ai]).pop() 
                      
#         # same for second fingerprint
#         if bi in processed: continue

#         if score > threshold:
#             # merge fingerprints
#             fp = updateFingerprint(afp, bfp, len(fmap[ai]), len(fmap[bi]))
#             merged_fps.append(fp)
#             # merge node references
#             i = len(merged_fps) - 1
#             merged_fmap[i] = list(set(fmap[ai] + fmap[bi]))
#             # mark as processed
#             processed += [ai, bi]
            
#     # add fingerprints that were not merged
#     for i, fp in enumerate(fps):
#         if i not in processed:
#             merged_fps.append(fp)
#             merged_fmap[len(merged_fps) - 1] = fmap[i]

#     return merged_fps, merged_fmap



