import fileinput
import os.path
import urllib

import sys
import re
import argparse
import contextlib
import time

import numpy as np
	
# definition of constants
_ERROR=-1
_SUCCESS=1
_DEBUG=0
_SMALLN=0.000000001

global _PRINTLEV

_PRINTLEV=5


def convertNAN(val):
	if val == np.nan:
		return 0 #replace NaN 
	return val


def is_better(val1, val2, how):
	if (how=='min' or how=='std' or how=='mean') and val1 < val2:
		return 1
	if (how=='max' or how=='k' or how=='sum') and val1 > val2:
		return 1
	return 0
	
def is_int(v):
    v = str(v).strip()
    return v=='0' or (v if v.find('..') > -1 else v.lstrip('-+').rstrip('0').rstrip('.')).isdigit()


def is_writable(dirname):
	if os.access(dirname, os.F_OK): #if storage folder exixts
		if _PRINTLEV >0:
			print("Storage folder exixts ")
		if os.access(dirname, os.W_OK):
			if _PRINTLEV >0:
				print("Storage folder is writable ")
			return _SUCCESS
		else:
			if _PRINTLEV >0:
				print("Storage folder is NOT writable ")
			return _ERROR
	else: # create it if doesn't exist
		if os.mkdir(dirname):
			if _PRINTLEV >0:
				print("Created storage folder")
			return _SUCCESS
		else:
			if _PRINTLEV >0:
				print("Cannot create storage folder")
			return _ERROR
			

def append_in_file(writepath, fname, line):
	file_to_write=writepath+'/k_'+fname
	mode = 'a' if os.path.exists(file_to_write) else 'w'
	with open(file_to_write, mode) as f:
		f.write(line)
		f.write('\n')
		return _SUCCESS
	return _ERROR
			

def lock_file(fname):
	#check that the file exixts
	print("print level is "+str(_PRINTLEV))
	if not os.path.isfile(fname):
		if _PRINTLEV >0:
			print("ERROR: file "+fname+" des not exist ")
		return _ERROR
		
	# check if the file is locked already
	print(fname+".lock")
	if os.path.isfile(fname+".lock"):
		if _PRINTLEV >0:
			print("ERROR: file "+fname+" is locked")
		return _ERROR
	else:
		# create the lock
		fname_lock = fname+".lock"
		file = open(fname_lock, 'w')
		file.close()
		if _PRINTLEV >1:
			print("file "+fname+" successfully locked")
		return _SUCCESS
		
		
def unlock_file(fname):
	# check if the file is locked already
	if os.path.isfile(fname+".lock"):
		if not os.remove(fname+".lock"):
			if _PRINTLEV >1:
				print("lock removed from "+fname)
			return _SUCCESS
		else:
			if _PRINTLEV >0:
				print("ERROR: unable to remove lock for "+fname)
			return _ERROR
	else:
		if _PRINTLEV >0:
			print("ERROR: lock not available for "+fname)
		return _ERROR
	
		
def replace_file(fnameold,fnamenew):
	if not os.rename(fnameold, fnamenew):
		if _PRINTLEV >1:
			print("file "+fnameold+" renamed to "+fnamenew)
		return _SUCCESS
		
	if _PRINTLEV >0:
		print("ERROR: unable to rename file")
	return _ERROR


def concatenate_files(outfilename,filenames):
	with open(outfilename, 'w') as fout:
	    for line in fileinput.input(filenames):
	        fout.write(line)
	

# get file from url
def get_remote_file(url_path):
	
	the_file=_ERROR
	req = urllib.request.Request(url_path)
	with contextlib.closing(urllib.urlopen(req)) as response:
		the_file = urllib.response.read()
	return the_file
	

def get_file(url_path):
	file = urllib.urlopen(url_path).read()

	


def save_url_message(url_file,url_path,model_name,data_name,last_trained,node_ip,sufix,cidx):
	# lock the file to save the message
	lock_file(url_file)
	
	# NOTE: model_name and data_name need the actual URL path, also url_file
	if lock_file(url_file)== _SUCCESS:
		file = open(url_file, 'w')
		file.write("{\n")
		file.write("    \"data_file\" : "+url_path+"/"+data_name+",\n")
		file.write("    \"model_file\" : "+url_path+"/"+model_name+",\n")
		file.write("    \"new_instances\" : "+str(last_trained)+",\n")
		file.write("    \"timestamp\" : "+sufix+",\n")
		file.write("    \"confidence\" : "+cidx+",\n")
		file.write("    \"node_ip\" : "+str(node_ip)+",\n")
		file.write("},\n")
		file.close()
		unlock_file(url_file)
		return _SUCCESS
		
	return _ERROR
	
	
def dump_data(dump_data,sufix):
	# just replace local and global files
	replace_file(dump_data,dump_data+sufix) # add a sufix based on time as unique id
	return dump_data+sufix
	
		
# end of functions ---------------------

class Ddict(dict):
	def __init__(self, default=None):
		self.default = default

	def __getitem__(self, key):
		if not self.has_key(key):
			self[key] = self.default()
		return dict.__getitem__(self, key)
# end of bidimensional dict