import sys
import re
import urllib
import argparse
import contextlib
import time
import json

global _PRINTLEV

from fpc.util import *
from fpc.metagraph import MetaGraph

# definition of constants
_ERROR=-1
_PRINTLEV=0
_SUCCESS=1
_DEBUG=0
_SMALLN=0.000000001


# PARAM constants==============================
_NODE_IP='127.0.0.1'
LOCAL_INPUT="working/my_input.txt"
LOCAL_MESSAGE="working/my_message.txt"
LOCAL_QUERY="working/my_query.txt"




data_flag=0
message_flag=0
query_flag=0

# input parameters
has_trained=0
last_trained=0
last_dump=0

rmsd=0
correct=0
incorrect=0



# parse arguments ------------------------
parser = argparse.ArgumentParser(description='This is a model generator daemon that computes distributed tree models ')
parser.add_argument('-f','--format_file',help='Format file with ARFF headers defining the format of the input', required=True)
parser.add_argument('-i','--input_file', help='Input file, containing m rows=number of elements, n-1 columns = number of attributes per element, and 1 aditional column containing the class of each element',required=True)
parser.add_argument('-q','--query_file', help='Query file, containing m rows=number of elements, n-1 columns = number of attributes per element',required=True)
parser.add_argument('-o','--output_file',help='Output file', required=True)
parser.add_argument('-m','--message_file',help='File to get information from the communication layer', required=True)
parser.add_argument('-u','--url_file',help='File to pass information to the communication layer', required=True)
parser.add_argument('-up','--url_path',help='Path to the URL where files will be published', required=True)
parser.add_argument('-dm','--dump_model',help='File to dump the current model + timestamp', required=True)
parser.add_argument('-dd','--dump_data',help='File to dump the current data + timestamp', required=True)
parser.add_argument('-save','--save_path',help='storage path for clustered files', required=True)
parser.add_argument('-p','--print_level',help='print(level from 0 to 5 defines the amount of messages that are printed', required=False)
parser.add_argument('-d','--debug', action='store_const', const=1, help='If this flag is present it will output information in debug mode',required=False)
parser.add_argument('-k','--kill',help='Defines number of empty iterations to kill the daemon', required=False)
parser.add_argument('-nf','--number_file',help='Defines the file number to begin processing', required=False)
parser.add_argument('-simC','--similarity_cluster',help='Defines the similarity metric to use in the clustering phase', required=False)
parser.add_argument('-simM','--similarity_merge',help='Defines the similarity metric to use in the merge', required=False)
parser.add_argument('-thC','--threshold_cluster',help='Defines the threshold to use during clustering', required=False)
parser.add_argument('-thM','--threshold_merge',help='Defines the threshold to use during merge', required=False)
parser.add_argument('-fnC','--function_cluster',help='Defines the function to use during clustering', required=False)
parser.add_argument('-fnM','--function_merge',help='Defines the function to use during merge', required=False)
				
args = parser.parse_args()




_DEBUG=args.debug
_PRINTLEV=int(args.print_level)

if _DEBUG:
	_PRINTLEV=6
	
	
if args.kill:
	killproc=int(args.kill)
else:
	killproc=-1
	
last_trained=0


if _PRINTLEV >3:
    print('Opening files {} {}'.format(args.output_file,args.format_file))

f = open(args.output_file, 'a');


fformat = open(args.format_file, 'r');

# first read the format file (json file) =======================================
with open(args.format_file, 'r') as jformat:  
	g = json.load(jformat)
	
	if _PRINTLEV > 3:
		print("Initializing graph ")
	MG = MetaGraph(g['graph'])
	
	if args.number_file: MG.number_file = args.number_file
	if args.similarity_cluster: MG.similarity_cluster = args.similarity_cluster
	if args.similarity_merge: MG.similarity_merge = args.similarity_merge
	if args.threshold_cluster: MG.threshold_cluster = float(args.threshold_cluster)
	if args.threshold_merge: MG.threshold_merge = float(args.threshold_merge)
	if args.function_cluster: MG.function_cluster = args.function_cluster
	if args.function_merge: MG.function_merge = args.function_merge
		
	MG._PRINTLEV = _PRINTLEV
	
	if _DEBUG or _PRINTLEV > 2:
		MG.print_fields()

	
if args.save_path:
	if not MG.check_storage(args.save_path):
		sys.exit("ERROR: It is not possible to save files in {}".format(arga.save_path))

	
# save the current time	
last_dump= time.time()


# daemon body!=================================================

iterations=0

while(1):
	
	iterations=iterations+1
	
	if iterations==killproc:
		print("EXIT")
		exit(0)
		
	# check for input file
	if _PRINTLEV>0:
		print("Lock-Replace-Unlock input file "+args.input_file)
	if lock_file(args.input_file)== _SUCCESS:
		if replace_file(args.input_file,LOCAL_INPUT) == _SUCCESS:
			data_flag=1;
		unlock_file(args.input_file)
	if data_flag==1:
		if _PRINTLEV>0:
			print("Processing input data")
		last_trained=MG.process_input(LOCAL_INPUT, args.output_file,args.dump_data,args.dump_model,args.url_file,args.url_path,_NODE_IP,last_trained,last_dump)
		data_flag=0
	else:
		if _PRINTLEV>0:
			print("Did NOT process input data")
		
		
		
	# check for query file
	if _PRINTLEV>0:
		print("Lock-Replace-Unlock query file "+args.query_file)
	if MG.trained>0 and lock_file(args.query_file)== _SUCCESS:
		if replace_file(args.query_file,LOCAL_QUERY) == _SUCCESS:
			query_flag=1;
		elif _PRINTLEV>0:
			print("ERROR replacing query file")
			print(args.query_file)
			print(LOCAL_QUERY)
			
		unlock_file(args.query_file)
	elif _PRINTLEV>0:
		print("ERROR Locking query file")
		
	if query_flag==1:
		if _PRINTLEV>0:
			print("Processing query ")
		MG.process_query(LOCAL_QUERY,args.output_file)==_SUCCESS
		query_flag=0
	else:
		if _PRINTLEV>0:
			print("Did NOT process query")
		
		
		
	# check for message file
	if _PRINTLEV>0:
		print("Lock-Replace-Unlock message file "+args.message_file)
	if lock_file(args.message_file)== _SUCCESS:
		if replace_file(args.message_file,LOCAL_MESSAGE) == _SUCCESS:
			message_flag=1;
		unlock_file(args.message_file)
	if message_flag==1:
		if _PRINTLEV>0:
			print("Processing message ")
		MG.process_message(LOCAL_MESSAGE,args.dump_model,args.dump_data)
		message_flag=0
	else:
		if _PRINTLEV>0:
			print("Did NOT process message")
	


	if _PRINTLEV>0:
		print("Going to sleep...")
		
	time.sleep(5)
	
	if _PRINTLEV>0:
		print("Waking up...")
		
		
		
		

#python3 graph_daemon.py -f ddata/amazon-ratings/amazon-ratings.json -i ddata/amazon-ratings/amazon-ratings.csv -o output.txt -q query.txt -m message.txt -u url.txt -up publish/ -dm publish/dumpmodel.txt -dd publish/dumpdata.txt -k 3 -p 5 -bs 20


